﻿using DiplomaThesis.Core.Encryption;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Numerics;

namespace DiplomaThesis.Core.Tests.Encryption;

/// <summary>
/// Class for testing diffie hellman implementation
/// </summary>
[TestClass]
public class DiffieHellmanTests
{
    /// <summary>
    /// Does diffie hellman key exchange and simulates sending of message 
    /// </summary>
    /// <param name="message">Message to be send</param>
    /// <param name="primeNumberSize">Size of the prime number used for diffie hellman key calculations</param>
    /// <returns>Decrypted message after diffie hellman key exchange and simulation of message sending</returns>
    private string DiffieHellmanSendMessage(string message, PrimeNumberSize primeNumberSize)
    {
        // Get pair of big prime number and generator
        var primeGeneratorPair = new PrimeGeneratorPairHelper().GetPrimeGeneratorPair(primeNumberSize);

        // Initialize big integer generator
        BigIntegerGenerator generator = new BigIntegerGenerator();

        // Generate person one and two secrets
        BigInteger personOneSecret = generator.GenerateBigInteger(64);
        BigInteger personTwoSecret = generator.GenerateBigInteger(64);

        // Initialize diffie hellman helper class
        DiffieHellman diffieHellman = new DiffieHellman();

        // Calculate public key of person one based on prime and generator pair
        BigInteger personOnePublicKey = diffieHellman.CalculatePublicKey(
            primeGeneratorPair.Prime,
            primeGeneratorPair.Generator,
            personOneSecret);

        // Calculate public key of person two based on prime and generator pair
        BigInteger personTwoPublicKey = diffieHellman.CalculatePublicKey(
            primeGeneratorPair.Prime,
            primeGeneratorPair.Generator,
            personTwoSecret);

        // Calculate common key from perspective of person one
        BigInteger person1CommonKey = diffieHellman.CalculateCommonKey(
            primeGeneratorPair.Prime,
            personTwoPublicKey,
            personOneSecret);

        // Calculate common key from perspective of person two
        BigInteger person2CommonKey = diffieHellman.CalculateCommonKey(
            primeGeneratorPair.Prime,
            personOnePublicKey,
            personTwoSecret);

        // Calculate SHA256 from common key of person one
        byte[] person1Hash = diffieHellman.HashCommonKey(person1CommonKey);

        // Calculate SHA256 from common key of person two
        byte[] person2Hash = diffieHellman.HashCommonKey(person2CommonKey);

        // Initialize class for encrypting messages
        SymmetricEncryptor encryptor = new SymmetricEncryptor();

        // Encrypt message from perspective of person one
        var encryptedPayload = encryptor.EncryptMessage(person1Hash, message);

        // Decrypt message from perspective of person two
        var decryptedMessage = encryptor.DecryptMessage(person2Hash, encryptedPayload.IV, encryptedPayload.EncryptedMessage);

        // Return decrypted messsage
        return decryptedMessage;
    }

    /// <summary>
    /// Tries to do diffie-hellman with 1536-bit long prime number
    /// </summary>
    [TestMethod]
    public void DiffieHellman1536()
    {
        // Message to be tested
        string message = "Hello world!";

        // Test if final message is equal to initial message
        Assert.AreEqual(
            message, 
            DiffieHellmanSendMessage(message, PrimeNumberSize.Size1536));
    }

    /// <summary>
    /// Tries to do diffie-hellman with 2048-bit long prime number
    /// </summary>
    [TestMethod]
    public void DiffieHellman2048()
    {
        // Message to be tested
        string message = "Hello world!";

        // Test if final message is equal to initial message
        Assert.AreEqual(
            message,
            DiffieHellmanSendMessage(message, PrimeNumberSize.Size2048));
    }

    /// <summary>
    /// Tries to do diffie-hellman with 3072-bit long prime number
    /// </summary>
    [TestMethod]
    public void DiffieHellman3072()
    {
        // Message to be tested
        string message = "Hello world!";

        // Test if final message is equal to initial message
        Assert.AreEqual(
            message,
            DiffieHellmanSendMessage(message, PrimeNumberSize.Size3072));
    }

    /// <summary>
    /// Tries to do diffie-hellman with 4096-bit long prime number
    /// </summary>
    [TestMethod]
    public void DiffieHellman4096()
    {
        // Message to be tested
        string message = "Hello world!";

        // Test if final message is equal to initial message
        Assert.AreEqual(
            message,
            DiffieHellmanSendMessage(message, PrimeNumberSize.Size4096));
    }

    /// <summary>
    /// Tries to do diffie-hellman with 6144-bit long prime number
    /// </summary>
    [TestMethod]
    public void DiffieHellman6144()
    {
        // Message to be tested
        string message = "Hello world!";

        // Test if final message is equal to initial message
        Assert.AreEqual(
            message,
            DiffieHellmanSendMessage(message, PrimeNumberSize.Size6144));
    }

    /// <summary>
    /// Tries to do diffie-hellman with 8192-bit long prime number
    /// </summary>
    [TestMethod]
    public void DiffieHellman8192()
    {
        // Message to be tested
        string message = "Hello world!";

        // Test if final message is equal to initial message
        Assert.AreEqual(
            message,
            DiffieHellmanSendMessage(message, PrimeNumberSize.Size8192));
    }
}
