﻿using System.Numerics;
using System.Security.Cryptography;

namespace DiplomaThesis.Core.Encryption;

/// <summary>
/// Diffie hellman implementation 
/// </summary>
public class DiffieHellman
{
    /// <summary>
    /// Calculates common key
    /// </summary>
    /// <param name="primeNumber">Prime number used for common key calculation</param>
    /// <param name="peerPublicKey">Public key of peer</param>
    /// <param name="secretKey">Secret key</param>
    /// <returns>Calculated common key</returns>
    public BigInteger CalculateCommonKey(
        BigInteger primeNumber,
        BigInteger peerPublicKey,
        BigInteger secretKey)
    {
        // Return calculated common key
        return (peerPublicKey ^ secretKey) % primeNumber;
    }

    /// <summary>
    /// Calculates public key
    /// </summary>
    /// <param name="primeNumber">Prime number used for public key calculation</param>
    /// <param name="generator">Generator used for public key calculation</param>
    /// <param name="secretKey">Secret key</param>
    /// <returns>Calculated public key</returns>
    public BigInteger CalculatePublicKey(
        BigInteger primeNumber,
        BigInteger generator,
        BigInteger secretKey)
    {
        // Return calculated public key
        return (generator ^ secretKey) % primeNumber;
    }

    /// <summary>
    /// Calculates SHA256 hash for common key
    /// </summary>
    /// <param name="commonKey">Common key for which hash will be calculated</param>
    /// <returns>Hash calculated from common key</returns>
    public byte[] HashCommonKey(BigInteger commonKey)
    {
        // Use SHA256 instance
        using var sha256 = SHA256.Create();

        // Return computed hash from common key
        return sha256.ComputeHash(commonKey.ToByteArray());
    }
}
