﻿using DiplomaThesis.MobileApp.Models.Api;
using DiplomaThesis.MobileApp.Models.ApiPayload.Error;
using System.Net.Http;
using System.Text;
using System.Text.Json;
using System.Threading.Tasks;

namespace DiplomaThesis.MobileApp.Helpers
{
    /// <summary>
    /// Helper for working with HTTP protocol
    /// </summary>
    public class HttpHelper
    {
        /// <summary>
        /// Client used for HTTP requests
        /// </summary>
        private readonly HttpClient mHttpClient;

        /// <summary>
        /// Default constructor
        /// </summary>
        public HttpHelper()
        {
            // Initialize HTTP client
            mHttpClient = new HttpClient();
        }

        /// <summary>
        /// Does GET request to specified url
        /// </summary>
        /// <typeparam name="TResponsePayload">Type of the response payload</typeparam>
        /// <param name="url">Url of the request</param>
        /// <returns>Response model which is result of GET request</returns>
        public async Task<TResponsePayload> Get<TResponsePayload>(string url)
        {
            // Do GET request on server
            var response = await mHttpClient.GetAsync(url);

            // Convert response to string
            var responseString = await response.Content.ReadAsStringAsync();

            // Deserialize JSON to model
            var deserializedModel = JsonSerializer.Deserialize<ApiResponse<object>>(
                responseString,
                new JsonSerializerOptions
                {
                    PropertyNameCaseInsensitive = true
                });

            // If response type is OK deserialize again with proper type
            if (deserializedModel.ResponseType == ApiResponseType.OK)
            {
                var deserializedModel2 = JsonSerializer.Deserialize<ApiResponse<TResponsePayload>>(
                    responseString,
                    new JsonSerializerOptions
                    {
                        PropertyNameCaseInsensitive = true
                    });

                return deserializedModel2.Payload;
            }
            else
            {
                throw new System.Exception("API error: " + ((ErrorPayload)deserializedModel.Payload).ErrorType.ToString());
            }
        }

        /// <summary>
        /// Does POST request to specified url
        /// </summary>
        /// <typeparam name="TRequestPayload">Type of the request payload</typeparam>
        /// <typeparam name="TResponsePayload">Type of the response payload</typeparam>
        /// <param name="request">Request which will be sent to the server</param>
        /// <param name="url">Url of the request</param>
        /// <returns>Response model which is result of POST request</returns>
        public async Task<TResponsePayload> Post<TRequestPayload, TResponsePayload>(
            ApiRequest<TRequestPayload> request,
            string url)
            where TResponsePayload : class
        {
            // Serialize input request to JSON
            var serializedRequest = JsonSerializer.Serialize<ApiRequest<TRequestPayload>>(request);

            // Create http content from JSON
            var stringContent = new StringContent(serializedRequest, Encoding.UTF8, "application/json");

            // Do POST request on server
            var response = await mHttpClient.PostAsync(url, stringContent);

            // Convert response to string
            var responseString = await response.Content.ReadAsStringAsync();

            // Deserialize JSON to response model
            var deserializedModel = JsonSerializer.Deserialize<ApiResponse<object>>(
                responseString,
                new JsonSerializerOptions
                {
                    PropertyNameCaseInsensitive = true
                });

            // If response type is OK deserialize again with proper type
            if (deserializedModel.ResponseType == ApiResponseType.OK)
            {
                var deserializedModel2 = JsonSerializer.Deserialize<ApiResponse<TResponsePayload>>(
                responseString,
                new JsonSerializerOptions
                {
                    PropertyNameCaseInsensitive = true
                });

                return deserializedModel2.Payload;
            }
            else
            {
                throw new System.Exception("API error: " + ((ErrorPayload)deserializedModel.Payload).ErrorType.ToString());
            }
        }

        /// <summary>
        /// Does PUT request to specified url
        /// </summary>
        /// <typeparam name="TRequestPayload">Type of the request payload</typeparam>
        /// <typeparam name="TResponsePayload">Type of the response payload</typeparam>
        /// <param name="request">Request which will be sent to the server</param>
        /// <param name="url">Url of the request</param>
        /// <returns>Response model which is result of POST request</returns>
        public async Task<TResponsePayload> Put<TRequestPayload, TResponsePayload>(
            ApiRequest<TRequestPayload> request,
            string url)
        {
            // Serialize input model to JSON
            var serializedModel = JsonSerializer.Serialize<ApiRequest<TRequestPayload>>(request);

            // Create http content from JSON model
            var stringContent = new StringContent(serializedModel, Encoding.UTF8, "application/json");

            // Do PUT request on server
            var response = await mHttpClient.PutAsync(url, stringContent);

            // Convert response to string
            var responseString = await response.Content.ReadAsStringAsync();

            // Deserialize JSON to model
            var deserializedModel = JsonSerializer.Deserialize<ApiResponse<object>>(
                responseString,
                new JsonSerializerOptions
                {
                    PropertyNameCaseInsensitive = true
                });

            // If response type is OK deserialize again with proper type
            if (deserializedModel.ResponseType == ApiResponseType.OK)
            {
                var deserializedModel2 = JsonSerializer.Deserialize<ApiResponse<TResponsePayload>>(
                    responseString,
                    new JsonSerializerOptions
                    {
                        PropertyNameCaseInsensitive = true
                    });

                return deserializedModel2.Payload;
            }
            else
            {
                throw new System.Exception("API error: " + ((ErrorPayload)deserializedModel.Payload).ErrorType.ToString());
            }
        }
    }
}
