﻿using DiplomaThesis.MobileApp.Encryption;
using DiplomaThesis.MobileApp.Helpers;
using DiplomaThesis.MobileApp.Models.Api;
using DiplomaThesis.MobileApp.Models.ApiPayload.Token;
using DiplomaThesis.MobileApp.Models.ApiPayload.User;
using System.Numerics;
using System.Threading.Tasks;

namespace DiplomaThesis.MobileApp.Services
{
    /// <summary>
    /// Service used for authorization on server
    /// </summary>
    public class AuthService
    {
        /// <summary>
        /// Big integer generator helper
        /// </summary>
        private readonly BigIntegerGenerator mBigIntegerGenerator;

        /// <summary>
        /// Diffie-Hellman helper
        /// </summary>
        private readonly DiffieHellman mDiffieHellman;

        /// <summary>
        /// Helper for working with HTTP
        /// </summary>
        private readonly HttpHelper mHttpHelper;

        /// <summary>
        /// Prime & Generator pair helper
        /// </summary>
        private readonly PrimeGeneratorPairHelper mPrimeGeneratorPairHelper;

        /// <summary>
        /// Encryptor which uses symmetric encryption
        /// </summary>
        private readonly SymmetricEncryptor mSymmetricEncryptor;

        /// <summary>
        /// Default constructor
        /// </summary>
        public AuthService()
        {
            mBigIntegerGenerator = new BigIntegerGenerator();
            mDiffieHellman = new DiffieHellman();
            mHttpHelper = new HttpHelper();
            mPrimeGeneratorPairHelper = new PrimeGeneratorPairHelper();
            mSymmetricEncryptor = new SymmetricEncryptor();
        }

        /// <summary>
        /// Registers user on server
        /// </summary>
        /// <param name="serverUrl">URL of the server</param>
        /// <param name="displayName">Display name of the user</param>
        /// <returns>Registration data</returns>
        public async Task<(string UserId, string EncryptionKey)> Register(string serverUrl, string displayName)
        {
            var secretKey = mBigIntegerGenerator.GenerateBigInteger(32);
            var primeGeneratorPair = mPrimeGeneratorPairHelper.GetPrimeGeneratorPair(PrimeNumberSize.Size8192);

            var publicKey = mDiffieHellman.CalculatePublicKey(
                primeGeneratorPair.Prime,
                primeGeneratorPair.Generator,
                secretKey);

            var request = new ApiRequest<UserAddRequestPayload>()
            {
                Payload = new UserAddRequestPayload()
                {
                    DisplayName = displayName,
                    Generator = primeGeneratorPair.Generator.ToString(),
                    PublicKey = publicKey.ToString(),
                    Prime = primeGeneratorPair.Prime.ToString(),
                }
            };

            var response = await mHttpHelper
                .Post<UserAddRequestPayload, UserAddResponsePayload>(
                    request,
                    serverUrl + "/user");

            var commonKey = mDiffieHellman
                .CalculateCommonKey(
                    primeGeneratorPair.Prime,
                    BigInteger.Parse(response.ServerPublicKey),
                    secretKey);

            return (response.UserId, commonKey.ToString());
        }

        /// <summary>
        /// Logs in user on the server
        /// </summary>
        /// <param name="serverUrl">URL of the server</param>
        /// <param name="userId">Id of the user</param>
        /// <param name="encryptionKey">Encryption key of the user</param>
        /// <returns>Login data</returns>
        public async Task<(string token, string displayName)> Login(string serverUrl, string userId, string encryptionKey)
        {
            var response = await mHttpHelper
                .Get<TokenGetResponsePayload>($"{serverUrl}/token/{userId}");

            var token = mSymmetricEncryptor.DecryptMessage(
                mDiffieHellman.HashCommonKey(BigInteger.Parse(encryptionKey)),
                BigInteger.Parse(response.IV).ToByteArray(),
                BigInteger.Parse(response.EncryptedToken).ToByteArray());

            var userInfo = await mHttpHelper
                .Get<UserGetResponsePayload>($"{serverUrl}/user/{userId}");

            return (token, userInfo.DisplayName);
        }
    }
}
