﻿using DiplomaThesis.MobileApp.Encryption;
using DiplomaThesis.MobileApp.Helpers;
using DiplomaThesis.MobileApp.Models.Api;
using DiplomaThesis.MobileApp.Models.ApiPayload.Inbox;
using DiplomaThesis.MobileApp.Models.ApiPayload.User;
using DiplomaThesis.MobileApp.Models.Database;
using DiplomaThesis.MobileApp.Models.Inbox;
using DiplomaThesis.MobileApp.Models.Messaging;
using System;
using System.Numerics;
using System.Text.Json;
using System.Threading.Tasks;

namespace DiplomaThesis.MobileApp.Services
{
    /// <summary>
    /// Service used for sending and receiving messages
    /// </summary>
    public class MessageService
    {
        /// <summary>
        /// Big integer generator helper class instance
        /// </summary>
        private readonly BigIntegerGenerator mBigIntegerGenerator;

        /// <summary>
        /// Diffie-Hellman helper
        /// </summary>
        private readonly DiffieHellman mDiffieHellman;

        /// <summary>
        /// Helper for working with HTTP
        /// </summary>
        private readonly HttpHelper mHttpHelper;

        /// <summary>
        /// Prime & Generator pair helper class instance
        /// </summary>
        private readonly PrimeGeneratorPairHelper mPrimeGeneratorPairHelper;

        /// <summary>
        /// Encryptor which uses symmetric encryption
        /// </summary>
        private readonly SymmetricEncryptor mSymmetricEncryptor;

        /// <summary>
        /// Default constructor
        /// </summary>
        public MessageService()
        {
            mBigIntegerGenerator = new BigIntegerGenerator();
            mDiffieHellman = new DiffieHellman();
            mHttpHelper = new HttpHelper();
            mPrimeGeneratorPairHelper = new PrimeGeneratorPairHelper();
            mSymmetricEncryptor = new SymmetricEncryptor();
        }

        /// <summary>
        /// Starts new chat
        /// </summary>
        /// <param name="displayName">User display name</param>
        /// <param name="server">Server URL</param>
        /// <param name="userId">User id</param>
        /// <param name="peerServer">Peer server URL</param>
        /// <param name="peerUserId">Peer user id</param>
        /// <returns>New chat</returns>
        public async Task<Chat> StartChat(
            string displayName,
            string server,
            string userId,
            string peerServer,
            string peerUserId)
        {
            var responseUserData = await mHttpHelper
                .Get<UserGetResponsePayload>(peerServer + "/user/" + peerUserId);

            var secretKey = mBigIntegerGenerator.GenerateBigInteger(32);
            var primeGeneratorPair = mPrimeGeneratorPairHelper.GetPrimeGeneratorPair(PrimeNumberSize.Size8192);

            var publicKey = mDiffieHellman.CalculatePublicKey(
                primeGeneratorPair.Prime,
                primeGeneratorPair.Generator,
                secretKey);

            var chat = new Chat()
            {
                ChatId = Guid.NewGuid().ToString(),
                IsActive = false,
                Name = responseUserData.DisplayName,
                PrivateKey = secretKey.ToString(),
                EncryptionKey = string.Empty,
                Prime = primeGeneratorPair.Prime.ToString(),
                PeerServer = peerServer,
                PeerUserId = peerUserId
            };

            var request = new ApiRequest<InboxMessageAddRequestPayload>()
            {
                AccessToken = string.Empty,
                Payload = new InboxMessageAddRequestPayload()
                {
                    ChatId = chat.ChatId,
                    IV = string.Empty,
                    MessageType = InboxMessageType.Initial,
                    ReceiverId = peerUserId,
                    Content = JsonSerializer.Serialize(new InitialMessage()
                    {
                        Generator = primeGeneratorPair.Generator.ToString(),
                        Prime = primeGeneratorPair.Prime.ToString(),
                        PublicKey = publicKey.ToString(),
                        DisplayName = displayName,
                        Server = server,
                        UserId = userId
                    })
                }
            };

            await mHttpHelper
                .Post<InboxMessageAddRequestPayload, InboxMessageAddResponsePayload>(
                    request, 
                    peerServer + "/inbox/");

            return chat;
        }

        /// <summary>
        /// Accepts chat
        /// </summary>
        /// <param name="chatId">Id of the chat</param>
        /// <param name="initialMessage">Initial message</param>
        /// <returns>Accepted chat</returns>
        public async Task<Chat> AcceptChat(
            string chatId,
            InitialMessage initialMessage)
        {
            var secretKey = mBigIntegerGenerator.GenerateBigInteger(32);

            var encryptionKey = mDiffieHellman.CalculateCommonKey(
                BigInteger.Parse(initialMessage.Prime),
                BigInteger.Parse(initialMessage.PublicKey),
                secretKey);

            var publicKey = mDiffieHellman.CalculatePublicKey(
                BigInteger.Parse(initialMessage.Prime),
                BigInteger.Parse(initialMessage.Generator),
                secretKey);

            var chat = new Chat()
            {
                IsActive = true,
                ChatId = chatId,
                EncryptionKey = encryptionKey.ToString(),
                Name = initialMessage.DisplayName,
                PeerServer = initialMessage.Server,
                PeerUserId = initialMessage.UserId,
                Prime = initialMessage.Prime,
                PrivateKey = secretKey.ToString()
            };

            var request = new ApiRequest<InboxMessageAddRequestPayload>()
            {
                AccessToken = string.Empty,
                Payload = new InboxMessageAddRequestPayload()
                {
                    ChatId = chat.ChatId,
                    IV = string.Empty,
                    MessageType = InboxMessageType.Accept,
                    ReceiverId = chat.PeerUserId,
                    Content = JsonSerializer.Serialize(new AcceptMessage()
                    {
                        PublicKey = publicKey.ToString()
                    })
                }
            };

            await mHttpHelper.Post<InboxMessageAddRequestPayload, InboxMessageAddResponsePayload>(
                request, 
                chat.PeerServer + "/inbox");

            return chat;
        }

        /// <summary>
        /// Completes chat
        /// </summary>
        /// <param name="chat">Chat to be completed</param>
        /// <param name="acceptMessage">Accept message</param>
        /// <returns>Completes chat</returns>
        public Chat CompleteChat(
            Chat chat,
            AcceptMessage acceptMessage)
        {
            var encryptionKey = mDiffieHellman.CalculateCommonKey(
                BigInteger.Parse(chat.Prime),
                BigInteger.Parse(acceptMessage.PublicKey),
                BigInteger.Parse(chat.PrivateKey));

            chat.EncryptionKey = encryptionKey.ToString();
            chat.IsActive = true;

            return chat;
        }

        /// <summary>
        /// Sends message
        /// </summary>
        /// <param name="peerServer">Server of the peer</param>
        /// <param name="peerUserId">Id of the peer</param>
        /// <param name="chatId">Id of the chat</param>
        /// <param name="message">Message to be send</param>
        /// <param name="encryptionKey">Key to be used for encryption</param>
        public async Task SendMessage(
            string peerServer,
            string peerUserId,
            string chatId,
            string message,
            string encryptionKey)
        {
            var commonKey = mDiffieHellman.HashCommonKey(BigInteger.Parse(encryptionKey));
            var encryptedMessage = mSymmetricEncryptor.EncryptMessage(commonKey, message);

            var request = new ApiRequest<InboxMessageAddRequestPayload>()
            {
                AccessToken = string.Empty,
                Payload = new InboxMessageAddRequestPayload()
                {
                    ChatId = chatId,
                    Content = new BigInteger(encryptedMessage.EncryptMessage).ToString(),
                    IV = new BigInteger(encryptedMessage.IV).ToString(),
                    MessageType = InboxMessageType.Standard,
                    ReceiverId = peerUserId
                }
            };

            await mHttpHelper.Post<InboxMessageAddRequestPayload, InboxMessageAddResponsePayload>(
                request, 
                peerServer + "/inbox");
        }

        /// <summary>
        /// Downloads messages from home server
        /// </summary>
        /// <param name="server">Server URL</param>
        /// <param name="accessToken">Access token</param>
        /// <returns>Downloaded messages</returns>
        public async Task<InboxMessageGetResponsePayload[]> DownloadMessages(
            string server,
            string accessToken)
        {
            return await mHttpHelper.Get<InboxMessageGetResponsePayload[]>(
                server + "/inbox/" + accessToken);
        }

        /// <summary>
        /// Decrypts messages
        /// </summary>
        /// <param name="content">Encrypted content</param>
        /// <param name="encryptionKey">Key to be used for decryption</param>
        /// <param name="iv">Initialization vector</param>
        /// <returns>Decrypted message</returns>
        public string DecryptMessage(
            string content,
            string encryptionKey,
            string iv)
        {
            return mSymmetricEncryptor.DecryptMessage(
                mDiffieHellman.HashCommonKey(BigInteger.Parse(encryptionKey)),
                BigInteger.Parse(iv).ToByteArray(),
                BigInteger.Parse(content).ToByteArray());
        }
    }
}
