﻿using DiplomaThesis.MobileApp.Helpers;
using DiplomaThesis.MobileApp.Models.Database;
using DiplomaThesis.MobileApp.Models.Inbox;
using DiplomaThesis.MobileApp.Models.Messaging;
using DiplomaThesis.MobileApp.Pages;
using DiplomaThesis.MobileApp.Services;
using DiplomaThesis.MobileApp.ViewModels.Base;
using LiteDB;
using System;
using System.Collections.ObjectModel;
using System.Linq;
using System.Windows.Input;
using Xamarin.Forms;

namespace DiplomaThesis.MobileApp.ViewModels
{
    /// <summary>
    /// View model for chats page
    /// </summary>
    public class ChatsPageViewModel : BaseViewModel
    {
        /// <summary>
        /// Message service used for working with messages
        /// </summary>
        private readonly MessageService mMessageService = new MessageService();

        /// <summary>
        /// Command for navigating to page for starting new chat
        /// </summary>
        public ICommand GoToAddNewChatPageCommand { get; }

        /// <summary>
        /// Command for syncing data from server
        /// </summary>
        public ICommand SyncCommand { get; }

        /// <summary>
        /// Command for opening new chat
        /// </summary>
        public ICommand OpenChatCommand { get; }

        /// <summary>
        /// User chats
        /// </summary>
        public ObservableCollection<Chat> Chats { get; set; } = new ObservableCollection<Chat>();

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="chatsPage">Page to which view model is bind to</param>
        public ChatsPageViewModel(ChatsPage chatsPage)
        {
            try
            {
                using (var db = new LiteDatabase(LiteDbHelper.GetDatabasePath("AppDb.db")))
                {
                    Chats = new ObservableCollection<Chat>(db.GetCollection<Chat>("chats")
                        .FindAll()
                        .ToList());
                }
            }
            catch (Exception ex)
            {
                chatsPage.DisplayAlert("Error", ex.Message, "OK");
            }

            SyncCommand = new Command(async () =>
            {
                try
                {
                    using (var db = new LiteDatabase(LiteDbHelper.GetDatabasePath("AppDb.db")))
                    {
                        var account = db
                            .GetCollection<Account>("accounts")
                            .FindAll()
                            .FirstOrDefault();

                        var messages = await mMessageService
                            .DownloadMessages(account.Server, account.AccessToken);

                        foreach (var message in messages)
                        {
                            if (message.MessageType == InboxMessageType.Initial)
                            {
                                var deserializedContent = System.Text.Json.JsonSerializer
                                    .Deserialize<InitialMessage>(message.Content);

                                var chat = await mMessageService.AcceptChat(message.ChatId, deserializedContent);

                                db.GetCollection<Chat>("chats").Insert(chat);

                                Chats.Add(chat);
                            }
                            if (message.MessageType == InboxMessageType.Accept)
                            {
                                var deserializedContent = System.Text.Json.JsonSerializer
                                    .Deserialize<AcceptMessage>(message.Content);

                                var chats = db.GetCollection<Chat>("chats");

                                var chat = chats.FindOne(x => x.ChatId == message.ChatId);

                                var completedChat = mMessageService.CompleteChat(chat, deserializedContent);

                                chats.Update(chat);

                            }
                            if (message.MessageType == InboxMessageType.Standard)
                            {
                                var chat = db
                                    .GetCollection<Chat>("chats")
                                    .FindOne(x => x.ChatId == message.ChatId);

                                var decryptedMessage = mMessageService
                                    .DecryptMessage(message.Content, chat.EncryptionKey, message.IV);

                                db.GetCollection<Message>("messages").Insert(new Message()
                                {
                                    ChatId = message.ChatId,
                                    Date = DateTime.Now,
                                    IsReceived = true,
                                    Text = decryptedMessage
                                });
                            }
                        }

                        Chats = new ObservableCollection<Chat>(db.GetCollection<Chat>("chats")
                            .FindAll()
                            .ToList());
                    }
                }
                catch (Exception ex)
                {
                    await chatsPage.DisplayAlert("Error", ex.Message, "OK");
                }
            });

            GoToAddNewChatPageCommand = new Command(async () =>
            {
                await Shell.Current.Navigation.PushAsync(new AddNewChatPage());

                using (var db = new LiteDatabase(LiteDbHelper.GetDatabasePath("AppDb.db")))
                {
                    Chats = new ObservableCollection<Chat>(db.GetCollection<Chat>("chats")
                        .FindAll()
                        .ToList());
                }
            });

            OpenChatCommand = new Command<Chat>(async (chat) =>
            {
                await Shell.Current.Navigation.PushAsync(new ChatPage(chat));
            });
        }
    }
}
