﻿using DiplomaThesis.Core.Api;
using DiplomaThesis.Core.ApiPayload.Error;
using DiplomaThesis.Core.ApiPayload.Inbox;
using DiplomaThesis.Server.Database;
using DiplomaThesis.Server.Models;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Http;
using Microsoft.EntityFrameworkCore;
using System.Linq;

namespace DiplomaThesis.Server.Routes;

/// <summary>
/// Inbox route helper class
/// </summary>
public static class InboxRoute
{
    /// <summary>
    /// Maps inbox route to web application
    /// </summary>
    /// <param name="app">Web application to which subroutes will be mapped to</param>
    /// <returns>Updated web application with mapped route</returns>
    public static WebApplication MapInboxRoute(this WebApplication app)
    {
        // GET for getting all messages currently in inbox
        app.MapGet("/inbox/{accessToken}", async (
            ApplicationDbContext dbContext,
            string accessToken) =>
        {
            // If access token wasn't provided
            if (accessToken == null)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.WrongOrMissingParameters,
                        Message = "Access token not provided"
                    }
                });
            }

            // Try to find user with provided access token
            var user = await dbContext.Users
                .Include(x => x.AccessToken)
                .Include(x => x.AuthInfo)
                .FirstOrDefaultAsync(x => x.AccessToken.Token == accessToken);

            // If token object wasn't found
            if (user == null)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.WrongOrMissingParameters,
                        Message = "Provided access token is not valid"
                    }
                });
            }

            // Get new messages in inbox
            var dbMessages = await dbContext.InboxMessages
                .Include(x => x.Receiver)
                .Where(x => x.Receiver.Id == user.Id)
                .ToArrayAsync();

            // Initialize array of messages for response
            var messages = new InboxMessageGetResponsePayload[
                dbMessages.Length];

            // Go through every message loaded from database
            for (int i = 0; i < dbMessages.Length; i++)
            {
                // Initialize message object in array and fill it with data
                messages[i] = new InboxMessageGetResponsePayload()
                {
                    ChatId = dbMessages[i].ChatId,
                    Content = dbMessages[i].Content,
                    MessageType = dbMessages[i].MessageType,
                    IV = dbMessages[i].IV
                };
            }

            // Remove messages from database
            dbContext.RemoveRange(dbMessages);
            dbContext.SaveChanges();

            // Initialize response
            var response = new ApiResponse<InboxMessageGetResponsePayload[]>()
            {
                ResponseType = ApiResponseType.OK,
                Payload = messages
            };

            // Return response with payload of inbox messages
            return Results.Ok(response);
        });

        // POST for adding message to inbox
        app.MapPost("/inbox", async (
            ApplicationDbContext dbContext,
            ApiRequest<InboxMessageAddRequestPayload> request) =>
        {
            // Unload payload from request
            var payload = request.Payload;

            // Try to find receiver in database
            var receiver = await dbContext
                .Users
                .FirstOrDefaultAsync(x => x.Id.ToString() == payload.ReceiverId);

            // If user with specified id wasn't found
            if (receiver == null)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.WrongOrMissingParameters,
                        Message = "Receiver with specified id wasn't found"
                    }
                });
            }

            // Initialize new message model for database
            var message = new InboxMessageModel()
            {
                ChatId = payload.ChatId,
                Content = payload.Content,
                MessageType = payload.MessageType,
                Receiver = receiver,
                IV = payload.IV
            };

            // Add new message to database
            var createdMessage = (await dbContext.AddAsync(message)).Entity;

            // Save changes to database
            await dbContext.SaveChangesAsync();

            // Initialize new response
            var response = new ApiResponse<InboxMessageAddResponsePayload>()
            {
                ResponseType = ApiResponseType.OK,
                Payload = new InboxMessageAddResponsePayload()
                {
                    ChatId = createdMessage.ChatId,
                    Content = createdMessage.Content,
                    InboxId = createdMessage.Id.ToString(),
                    MessageType = createdMessage.MessageType,
                    ReceiverId = createdMessage.Receiver.Id.ToString(),
                    IV = createdMessage.IV
                }
            };

            // Return response with payload of new inbox message
            return Results.Ok(response);
        });

        // Return web application
        return app;
    }
}
