﻿using DiplomaThesis.Core.Api;
using DiplomaThesis.Core.ApiPayload.Error;
using DiplomaThesis.Core.ApiPayload.Token;
using DiplomaThesis.Core.Encryption;
using DiplomaThesis.Server.Database;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Http;
using Microsoft.EntityFrameworkCore;
using System;
using System.Numerics;

namespace DiplomaThesis.Server.Routes;

/// <summary>
/// Token route helper class
/// </summary>
public static class TokenRoute
{
    /// <summary>
    /// Maps token route to web application
    /// </summary>
    /// <param name="app">Web application to which subroutes will be mapped to</param>
    /// <returns>Updated web application with mapped route</returns>
    public static WebApplication MapTokenRoute(this WebApplication app)
    {
        // GET for getting token by user id
        app.MapGet("/token/{userId}", async (
            ApplicationDbContext dbContext,
            DiffieHellman diffieHellman,
            SymmetricEncryptor encryptor,
            string userId) =>
        {
            // If id wasn't provided..
            if (userId == null)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.WrongOrMissingParameters,
                        Message = "Id of the user not provided"
                    }
                });
            }

            // Try to find user based on provided id
            var user = await dbContext.Users
                .Include(x => x.AccessToken)
                .Include(x => x.AuthInfo)
                .FirstOrDefaultAsync(x => x.Id.ToString() == userId);

            // If user with provided id wasn't found...
            if (user == null)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.WrongOrMissingParameters,
                        Message = "User with specified id wasn't found"
                    }
                });
            }

            // Get common key from database
            var commonKey = user.AuthInfo.CommonKey;

            // Hash the common key
            var hashedCommonKey = diffieHellman.HashCommonKey(BigInteger.Parse(commonKey));

            // Encrypt access token
            var encryptedTokenInfo = encryptor.EncryptMessage(hashedCommonKey, user.AccessToken.Token);

            // Initialize response
            var response = new ApiResponse<TokenGetResponsePayload>()
            {
                ResponseType = ApiResponseType.OK,
                Payload = new TokenGetResponsePayload()
                {
                    EncryptedToken = new BigInteger(encryptedTokenInfo.EncryptedMessage).ToString(),
                    IV = new BigInteger(encryptedTokenInfo.IV).ToString(),
                }
            };

            // Return response with payload of encrypted token data
            return Results.Ok(response);
        });

        // GET for getting access token reset
        app.MapGet("/token/reset/{userId}", async (
            ApplicationDbContext dbContext,
            DiffieHellman diffieHellman,
            SymmetricEncryptor encryptor,
            string userId) =>
        {
            // If id wasn't provided..
            if (userId == null)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.WrongOrMissingParameters,
                        Message = "Id of the user not provided"
                    }
                });
            }

            // Try to find user based on provided id
            var user = await dbContext.Users
                .Include(x => x.AccessToken)
                .Include(x => x.AuthInfo)
                .FirstOrDefaultAsync(x => x.Id.ToString() == userId);

            // If user with provided id wasn't found...
            if (user == null)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.WrongOrMissingParameters,
                        Message = "User with specified id wasn't found"
                    }
                });
            }

            // Update user with new access token
            user = user with
            {
                AccessToken = user.AccessToken with
                {
                    Token = Guid.NewGuid().ToString(),
                }
            };

            // Update user data in database
            var updatedUser = dbContext.Update(user).Entity;

            // Save changes to database
            await dbContext.SaveChangesAsync();

            // Get common key from database
            var commonKey = user.AuthInfo.CommonKey;

            // Hash the common key
            var hashedCommonKey = diffieHellman.HashCommonKey(BigInteger.Parse(commonKey));

            // Encrypt access token
            var encryptedTokenInfo = encryptor.EncryptMessage(hashedCommonKey, user.AccessToken.Token);

            // Initialize response
            var response = new ApiResponse<TokenGetResponsePayload>()
            {
                ResponseType = ApiResponseType.OK,
                Payload = new TokenGetResponsePayload()
                {
                    EncryptedToken = new BigInteger(encryptedTokenInfo.EncryptedMessage).ToString(),
                    IV = new BigInteger(encryptedTokenInfo.IV).ToString(),
                }
            };

            // Return response with payload of encrypted token data
            return Results.Ok(response);
        });

        // Return web application
        return app;
    }
}
