﻿using DiplomaThesis.Core.Api;
using DiplomaThesis.Core.ApiPayload.Error;
using DiplomaThesis.Core.ApiPayload.User;
using DiplomaThesis.Core.Encryption;
using DiplomaThesis.Server.Database;
using DiplomaThesis.Server.Models;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Http;
using Microsoft.EntityFrameworkCore;
using System;
using System.Numerics;

namespace DiplomaThesis.Server.Routes;

/// <summary>
/// User route helper class
/// </summary>
public static class UserRoute
{
    /// <summary>
    /// Maps user route to web application
    /// </summary>
    /// <param name="app">Web application to which subroutes will be mapped to</param>
    /// <returns>Updated web application with mapped route</returns>
    public static WebApplication MapUserRoute(this WebApplication app)
    {
        // GET for getting user by id
        app.MapGet("/user/{id}", async ( 
            ApplicationDbContext dbContext,
            string id) =>
        {
            // If id wasn't provided...
            if (id == null)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.WrongOrMissingParameters,
                        Message = "Id of the user not provided"
                    }
                });
            }

            // Try to find user with provided id
            var user = await dbContext
                .Users
                .FirstOrDefaultAsync(x => x.Id.ToString() == id);

            // If user with specified id wasn't found
            if (user == null)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.WrongOrMissingParameters,
                        Message = "User with specified id wasn't found"
                    }
                });
            }

            // Initialize response
            var response = new ApiResponse<UserGetResponsePayload>()
            {
                ResponseType = ApiResponseType.OK,
                Payload = new UserGetResponsePayload()
                {
                    Id = user.Id.ToString(),
                    DisplayName = user.DisplayName
                }
            };

            // Return api response with payload of user data
            return Results.Ok(response);
        });

        // POST for creating new user in database
        app.MapPost("/user", async (
            ApplicationDbContext dbContext,
            BigIntegerGenerator bigIntegerGenerator,
            DiffieHellman diffieHellman,
            SymmetricEncryptor encryptor,
            ApiRequest<UserAddRequestPayload> request) =>
        {
            // Try validate request 
            if (!MiniValidation.MiniValidator.TryValidate(request, out _))
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.RequestStructureNotValid,
                        Message = "Request structure is not valid"
                    }
                });
            }

            // Try validate request payload 
            if (!MiniValidation.MiniValidator.TryValidate(request.Payload, out _))
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.RequestStructureNotValid,
                        Message = "Payload structure is not valid"
                    }
                });
            }

            // Unload the payload
            var requestPayload = request.Payload;

            // Initialize server private key
            BigInteger serverPrivateKey = bigIntegerGenerator.GenerateBigInteger(32);

            // Initialize server public key
            BigInteger serverPublicKey = diffieHellman.CalculatePublicKey(
                    BigInteger.Parse(requestPayload.Prime),
                    BigInteger.Parse(requestPayload.Generator),
                    serverPrivateKey);

            // Calculate common key
            BigInteger commonKey = diffieHellman.CalculateCommonKey(
                BigInteger.Parse(requestPayload.Prime),
                BigInteger.Parse(requestPayload.PublicKey),
                serverPrivateKey);

            // Initialize new user model for database
            var user = new UserModel()
            {
                DisplayName = requestPayload.DisplayName,
                AccessToken = new AccessTokenModel()
                {
                    Token = Guid.NewGuid().ToString()
                },
                AuthInfo = new AuthInfoModel()
                {
                    CommonKey = commonKey.ToString(),
                    ServerPrivateKey = serverPrivateKey.ToString(),
                }
            };

            // Add user to the database
            var createdUser = (await dbContext.AddAsync(user)).Entity;

            // Save changes to database
            await dbContext.SaveChangesAsync();

            // Initialize response
            var response = new ApiResponse<UserAddResponsePayload>()
            {
                ResponseType = ApiResponseType.OK,
                Payload = new UserAddResponsePayload()
                {
                    ServerPublicKey = serverPublicKey.ToString(),
                    UserId = createdUser.Id.ToString(),
                    DisplayName = createdUser.DisplayName
                }
            };

            // Return api response with payload of newly created user data
            return Results.Ok(response);
        });

        // PUT for updating existing user in database 
        app.MapPut("/user", async (
            ApplicationDbContext dbContext,
            ApiRequest<UserUpdateRequestPayload> request) =>
        {
            // Try validate request 
            if (!MiniValidation.MiniValidator.TryValidate(request, out _))
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.RequestStructureNotValid,
                        Message = "Request structure is not valid"
                    }
                });
            }

            // Try validate request payload 
            if (!MiniValidation.MiniValidator.TryValidate(request.Payload, out _))
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.RequestStructureNotValid,
                        Message = "Payload structure is not valid"
                    }
                });
            }

            // Unload payload from request
            var payload = request.Payload;

            // Try to find user with selected id
            var user = await dbContext.Users
                .Include(x => x.AccessToken)
                .Include(x => x.AuthInfo)
                .FirstOrDefaultAsync(x => x.Id.ToString() == payload.Id);

            // If none user was found...
            if (user == null)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.WrongOrMissingParameters,
                        Message = "User not found"
                    }
                });
            }

            // If provided access token isn't correct...
            if (request.AccessToken != user.AccessToken.Token)
            {
                // Return API response with error
                return Results.BadRequest(new ApiResponse<ErrorPayload>()
                {
                    ResponseType = ApiResponseType.Error,
                    Payload = new ErrorPayload()
                    {
                        ErrorType = ErrorPayloadType.AccessDenied,
                        Message = "Provided access token isn't correct"
                    }
                });
            }

            // Update user with data from payload
            user = user with
            {
                DisplayName = payload.DisplayName
            };

            // Update user data in database
            var updatedUser = dbContext.Update(user).Entity;

            // Save changes to database
            await dbContext.SaveChangesAsync();

            // Initialize response
            var response = new ApiResponse<UserUpdateResponsePayload>()
            {
                ResponseType = ApiResponseType.OK,
                Payload = new UserUpdateResponsePayload()
                {
                    DisplayName = updatedUser.DisplayName,
                    Id = updatedUser.Id.ToString()
                }
            };

            // Return api response with payload of updated user data
            return Results.Ok(response);
        });

        // Return web application
        return app;
    }
}
